# -*- coding: utf-8 -*-

"""News syncing, merging, splitting, and uploading."""

import argparse
import logging
import os.path
import sys

from datetime import timedelta, timezone
from pathlib import Path
from shutil import rmtree
from subprocess import run
from time import sleep

from pytility import parse_date

from .merge import merge_files
from .split import split_files
from .utils import date_from_file, now

LOGGER = logging.getLogger(__name__)
BASE_DIR = Path(__file__).resolve().parent.parent


def update_news(
    s3_src, path_feeds, path_merged, path_split, s3_dst, split_size=None, log_level=None
):
    """News syncing, merging, splitting, and uploading."""

    path_feeds = Path(path_feeds).resolve()
    path_merged = Path(path_merged).resolve()
    path_split = Path(path_split).resolve()

    LOGGER.info(
        "Sync from <%s>, merge from <%s> into <%s>, split into <%s>, upload to <%s>",
        s3_src,
        path_feeds,
        path_merged,
        path_split,
        s3_dst,
    )

    LOGGER.info("Deleting existing dir <%s>", path_split.parent)
    rmtree(path_split.parent, ignore_errors=True)

    path_feeds.mkdir(parents=True, exist_ok=True)
    path_merged.parent.mkdir(parents=True, exist_ok=True)
    path_split.parent.mkdir(parents=True, exist_ok=True)

    LOGGER.info("S3 sync from <%s> to <%s>", s3_src, path_feeds)
    run(["aws", "s3", "sync", s3_src, os.path.join(path_feeds, "")], check=True)

    merge_files(
        in_paths=path_feeds.rglob("*.jl"),
        out_path=path_merged,
        keys="article_id",
        key_types="string",
        latest=("published_at", "scraped_at"),
        latest_types=("date", "date"),
        latest_required=True,
        sort_latest=True,
        sort_descending=True,
        concat_output=True,
        log_level=log_level,
    )

    split_files(
        path_in=path_merged, path_out=path_split, size=split_size, exclude_empty=True
    )

    LOGGER.info("S3 sync from <%s> to <%s>", path_split.parent, s3_dst)
    run(
        [
            "aws",
            "s3",
            "sync",
            "--acl",
            "public-read",
            "--exclude",
            ".gitignore",
            "--exclude",
            ".DS_Store",
            "--exclude",
            ".bucket",
            "--size-only",
            "--delete",
            os.path.join(path_split.parent, ""),
            s3_dst,
        ],
        check=True,
    )

    LOGGER.info("Done updating news.")


def _parse_args():
    parser = argparse.ArgumentParser(
        description="News syncing, merging, splitting, and uploading."
    )
    parser.add_argument(
        "--src-bucket",
        "-b",
        default="scrape.news.recommend.games",
        help="S3 bucket with scraped data",
    )
    parser.add_argument(
        "--dst-bucket",
        "-B",
        default="news.recommend.games",
        help="S3 bucket to upload to",
    )
    parser.add_argument(
        "--feeds", "-f", default=BASE_DIR / "feeds" / "news", help="Scraped items"
    )
    parser.add_argument(
        "--merged",
        "-m",
        default=BASE_DIR / "feeds" / "news_merged.jl",
        help="Merged file",
    )
    parser.add_argument(
        "--split",
        "-s",
        default=BASE_DIR / "feeds" / "news_hosting" / "news_{number:05d}.json",
        help="Split file template",
    )
    parser.add_argument(
        "--split-size",
        "-S",
        type=int,
        default=25,
        help="number of items in each result file",
    )
    parser.add_argument(
        "--dont-run-before", "-d", help="Either a date or a file with date information"
    )
    parser.add_argument(
        "--interval",
        "-i",
        type=int,
        default=10 * 60,  # 10 minutes
        help="number of seconds to wait before next run",
    )
    parser.add_argument(
        "--verbose",
        "-v",
        action="count",
        default=0,
        help="log level (repeat for more verbosity)",
    )

    return parser.parse_args()


def main():
    """Command line entry point."""

    args = _parse_args()

    logging.basicConfig(
        stream=sys.stderr,
        level=logging.DEBUG if args.verbose > 0 else logging.INFO,
        format="%(asctime)s %(levelname)-8.8s [%(name)s:%(lineno)s] %(message)s",
    )

    LOGGER.info(args)

    dont_run_before = parse_date(
        args.dont_run_before, tzinfo=timezone.utc
    ) or date_from_file(args.dont_run_before, tzinfo=timezone.utc)

    if dont_run_before:
        LOGGER.info("Don't run before %s", dont_run_before.isoformat())
        sleep_seconds = dont_run_before.timestamp() - now().timestamp()
        if sleep_seconds > 0:
            LOGGER.info("Going to sleep for %.1f seconds", sleep_seconds)
            sleep(sleep_seconds)

    if args.interval and args.dont_run_before and not parse_date(args.dont_run_before):
        dont_run_before = now() + timedelta(seconds=args.interval)
        LOGGER.info(
            "Don't run next time before %s, writing tag to <%s>",
            dont_run_before.isoformat(),
            args.dont_run_before,
        )
        with open(args.dont_run_before, "w") as file_obj:
            file_obj.write(dont_run_before.isoformat())

    update_news(
        s3_src=f"s3://{args.src_bucket}/",
        path_feeds=args.feeds,
        path_merged=args.merged,
        path_split=args.split,
        s3_dst=f"s3://{args.dst_bucket}/",
        split_size=args.split_size,
        log_level="DEBUG"
        if args.verbose > 1
        else "INFO"
        if args.verbose > 0
        else "WARN",
    )


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

""" Scrapy items """

import logging

from datetime import date, datetime, timezone
from functools import partial

from pytility import (
    clear_list,
    normalize_space,
    parse_bool,
    parse_date,
    parse_float,
    parse_int,
)
from scrapy import Field
from scrapy.loader.processors import Identity, MapCompose
from scrapy.utils.project import get_project_settings
from scrapy_extensions import TypedItem
from w3lib.html import remove_tags

from .utils import (
    identity,
    now,
    parse_json,
    replace_all_entities,
    serialize_date,
    serialize_json,
    validate_range,
    validate_url,
)

IDENTITY = Identity()
LOGGER = logging.getLogger(__name__)
SETTINGS = get_project_settings()
POS_INT_PROCESSOR = MapCompose(
    identity,
    str,
    remove_tags,
    replace_all_entities,
    normalize_space,
    parse_int,
    partial(validate_range, lower=1),
)
NN_INT_PROCESSOR = MapCompose(
    identity,
    str,
    remove_tags,
    replace_all_entities,
    normalize_space,
    parse_int,
    partial(validate_range, lower=0),
)
POS_FLOAT_PROCESSOR = MapCompose(
    identity,
    str,
    remove_tags,
    replace_all_entities,
    normalize_space,
    parse_float,
    partial(validate_range, lower=0),
    lambda v: v or None,
)
NN_FLOAT_PROCESSOR = MapCompose(
    identity,
    str,
    remove_tags,
    replace_all_entities,
    normalize_space,
    parse_float,
    partial(validate_range, lower=0),
)
DATE_PROCESSOR = MapCompose(partial(parse_date, tzinfo=timezone.utc))
URL_PROCESSOR = MapCompose(
    IDENTITY, str, partial(validate_url, schemes=frozenset(("http", "https")))
)


def _clear_list(items):
    return clear_list(items) or None


def _json_output():
    return SETTINGS.get("FEED_FORMAT") in ("jl", "json", "jsonl", "jsonlines")


def _serialize_bool(item):
    return int(item) if isinstance(item, bool) else None


class GameItem(TypedItem):
    """ item representing a game """

    JSON_OUTPUT = SETTINGS.get("FEED_FORMAT") in ("jl", "json", "jsonl", "jsonlines")
    JSON_SERIALIZER = identity if JSON_OUTPUT else serialize_json
    BOOL_SERIALIZER = identity if JSON_OUTPUT else _serialize_bool

    name = Field(dtype=str, required=True)
    alt_name = Field(
        dtype=list,
        output_processor=_clear_list,
        serializer=JSON_SERIALIZER,
        parser=parse_json,
    )
    year = Field(
        dtype=int,
        dtype_convert=parse_int,
        input_processor=MapCompose(
            identity,
            str,
            remove_tags,
            replace_all_entities,
            normalize_space,
            parse_int,
            partial(validate_range, lower=-4000, upper=date.today().year + 10),
            lambda year: year or None,
        ),
        default=None,
    )
    game_type = Field(
        dtype=list,
        output_processor=_clear_list,
        serializer=JSON_SERIALIZER,
        parser=parse_json,
    )
    description = Field(
        dtype=str,
        input_processor=MapCompose(
            identity,
            str,
            remove_tags,
            replace_all_entities,
            partial(normalize_space, preserve_newline=True),
        ),
    )

    designer = Field(
        dtype=list,
        output_processor=_clear_list,
        serializer=JSON_SERIALIZER,
        parser=parse_json,
    )
    artist = Field(
        dtype=list,
        output_processor=_clear_list,
        serializer=JSON_SERIALIZER,
        parser=parse_json,
    )
    publisher = Field(
        dtype=list,
        output_processor=_clear_list,
        serializer=JSON_SERIALIZER,
        parser=parse_json,
    )

    url = Field(dtype=str, input_processor=URL_PROCESSOR)
    official_url = Field(
        dtype=list,
        input_processor=URL_PROCESSOR,
        output_processor=_clear_list,
        serializer=JSON_SERIALIZER,
        parser=parse_json,
    )
    image_url = Field(
        dtype=list,
        input_processor=URL_PROCESSOR,
        output_processor=_clear_list,
        serializer=JSON_SERIALIZER,
        parser=parse_json,
    )
    image_file = Field(serializer=JSON_SERIALIZER, parser=parse_json)
    video_url = Field(
        dtype=list,
        input_processor=URL_PROCESSOR,
        output_processor=_clear_list,
        serializer=JSON_SERIALIZER,
        parser=parse_json,
    )
    rules_url = Field(
        dtype=list,
        input_processor=URL_PROCESSOR,
        output_processor=_clear_list,
        serializer=JSON_SERIALIZER,
        parser=parse_json,
    )
    rules_file = Field(serializer=JSON_SERIALIZER, parser=parse_json)
    review_url = Field(
        dtype=list,
        input_processor=URL_PROCESSOR,
        output_processor=_clear_list,
        serializer=JSON_SERIALIZER,
        parser=parse_json,
    )
    external_link = Field(
        dtype=list,
        input_processor=URL_PROCESSOR,
        output_processor=_clear_list,
        serializer=JSON_SERIALIZER,
        parser=parse_json,
    )
    list_price = Field(dtype=str)  # currency?

    min_players = Field(
        dtype=int,
        dtype_convert=parse_int,
        input_processor=POS_INT_PROCESSOR,
        default=None,
    )
    max_players = Field(
        dtype=int,
        dtype_convert=parse_int,
        input_processor=POS_INT_PROCESSOR,
        default=None,
    )
    min_players_rec = Field(
        dtype=int,
        dtype_convert=parse_int,
        input_processor=POS_INT_PROCESSOR,
        default=None,
    )
    max_players_rec = Field(
        dtype=int,
        dtype_convert=parse_int,
        input_processor=POS_INT_PROCESSOR,
        default=None,
    )
    min_players_best = Field(
        dtype=int,
        dtype_convert=parse_int,
        input_processor=POS_INT_PROCESSOR,
        default=None,
    )
    max_players_best = Field(
        dtype=int,
        dtype_convert=parse_int,
        input_processor=POS_INT_PROCESSOR,
        default=None,
    )
    min_age = Field(
        dtype=int,
        dtype_convert=parse_int,
        input_processor=POS_INT_PROCESSOR,
        default=None,
    )
    max_age = Field(
        dtype=int,
        dtype_convert=parse_int,
        input_processor=POS_INT_PROCESSOR,
        default=None,
    )
    min_age_rec = Field(
        dtype=float,
        dtype_convert=parse_float,
        input_processor=POS_FLOAT_PROCESSOR,
        default=None,
    )
    max_age_rec = Field(
        dtype=float,
        dtype_convert=parse_float,
        input_processor=POS_FLOAT_PROCESSOR,
        default=None,
    )
    min_time = Field(
        dtype=int,
        dtype_convert=parse_int,
        input_processor=POS_INT_PROCESSOR,
        default=None,
    )
    max_time = Field(
        dtype=int,
        dtype_convert=parse_int,
        input_processor=POS_INT_PROCESSOR,
        default=None,
    )

    category = Field(
        dtype=list,
        output_processor=_clear_list,
        serializer=JSON_SERIALIZER,
        parser=parse_json,
    )
    mechanic = Field(
        dtype=list,
        output_processor=_clear_list,
        serializer=JSON_SERIALIZER,
        parser=parse_json,
    )
    cooperative = Field(
        dtype=bool,
        default=None,
        input_processor=IDENTITY,
        serializer=BOOL_SERIALIZER,
        parser=parse_bool,
    )
    compilation = Field(
        dtype=bool,
        default=None,
        input_processor=IDENTITY,
        serializer=BOOL_SERIALIZER,
        parser=parse_bool,
    )
    compilation_of = Field(
        dtype=list,
        input_processor=MapCompose(parse_int),
        output_processor=_clear_list,
        serializer=JSON_SERIALIZER,
        parser=parse_json,
    )
    family = Field(
        dtype=list,
        output_processor=_clear_list,
        serializer=JSON_SERIALIZER,
        parser=parse_json,
    )
    expansion = Field(
        dtype=list,
        output_processor=_clear_list,
        serializer=JSON_SERIALIZER,
        parser=parse_json,
    )
    implementation = Field(
        dtype=list,
        input_processor=MapCompose(parse_int),
        output_processor=_clear_list,
        serializer=JSON_SERIALIZER,
        parser=parse_json,
    )
    integration = Field(
        dtype=list,
        input_processor=MapCompose(parse_int),
        output_processor=_clear_list,
        serializer=JSON_SERIALIZER,
        parser=parse_json,
    )

    rank = Field(
        dtype=int,
        dtype_convert=parse_int,
        input_processor=POS_INT_PROCESSOR,
        default=None,
    )
    num_votes = Field(
        dtype=int,
        dtype_convert=parse_int,
        input_processor=NN_INT_PROCESSOR,
        default=None,
    )
    avg_rating = Field(
        dtype=float,
        dtype_convert=parse_float,
        input_processor=POS_FLOAT_PROCESSOR,
        default=None,
    )
    stddev_rating = Field(
        dtype=float,
        dtype_convert=parse_float,
        input_processor=NN_FLOAT_PROCESSOR,
        default=None,
    )
    bayes_rating = Field(
        dtype=float,
        dtype_convert=parse_float,
        input_processor=POS_FLOAT_PROCESSOR,
        default=None,
    )
    worst_rating = Field(
        dtype=int,
        dtype_convert=parse_int,
        input_processor=POS_INT_PROCESSOR,
        default=None,
    )
    best_rating = Field(
        dtype=int,
        dtype_convert=parse_int,
        input_processor=POS_INT_PROCESSOR,
        default=None,
    )

    complexity = Field(
        dtype=float,
        dtype_convert=parse_float,
        input_processor=POS_FLOAT_PROCESSOR,
        default=None,
    )
    easiest_complexity = Field(
        dtype=int,
        dtype_convert=parse_int,
        input_processor=POS_INT_PROCESSOR,
        default=None,
    )
    hardest_complexity = Field(
        dtype=int,
        dtype_convert=parse_int,
        input_processor=POS_INT_PROCESSOR,
        default=None,
    )
    language_dependency = Field(
        dtype=float,
        dtype_convert=parse_float,
        input_processor=POS_FLOAT_PROCESSOR,
        default=None,
    )
    lowest_language_dependency = Field(
        dtype=int,
        dtype_convert=parse_int,
        input_processor=POS_INT_PROCESSOR,
        default=None,
    )
    highest_language_dependency = Field(
        dtype=int,
        dtype_convert=parse_int,
        input_processor=POS_INT_PROCESSOR,
        default=None,
    )

    bgg_id = Field(
        dtype=int,
        dtype_convert=parse_int,
        input_processor=POS_INT_PROCESSOR,
        default=None,
    )
    freebase_id = Field(dtype=str)
    wikidata_id = Field(dtype=str)
    wikipedia_id = Field(dtype=str)
    dbpedia_id = Field(dtype=str)
    luding_id = Field(
        dtype=int,
        dtype_convert=parse_int,
        input_processor=POS_INT_PROCESSOR,
        default=None,
    )
    spielen_id = Field(dtype=str)
    bga_id = Field(dtype=str)

    published_at = Field(
        dtype=datetime,
        input_processor=DATE_PROCESSOR,
        serializer=serialize_date,
        parser=parse_date,
    )
    updated_at = Field(
        dtype=datetime,
        input_processor=DATE_PROCESSOR,
        serializer=serialize_date,
        parser=parse_date,
    )
    scraped_at = Field(
        dtype=datetime,
        required=True,
        default=now,
        input_processor=DATE_PROCESSOR,
        serializer=serialize_date,
        parser=parse_date,
    )


class UserItem(TypedItem):
    """ item representing a user """

    JSON_OUTPUT = SETTINGS.get("FEED_FORMAT") in ("jl", "json", "jsonl", "jsonlines")
    JSON_SERIALIZER = identity if JSON_OUTPUT else serialize_json

    item_id = Field(
        dtype=int,
        dtype_convert=parse_int,
        input_processor=POS_INT_PROCESSOR,
    )

    bgg_user_name = Field(
        dtype=str, required=True, input_processor=MapCompose(identity, str, str.lower)
    )
    first_name = Field(dtype=str)
    last_name = Field(dtype=str)

    registered = Field(
        dtype=int,
        dtype_convert=parse_int,
        input_processor=MapCompose(
            identity,
            parse_int,
            partial(validate_range, lower=1999, upper=date.today().year),
        ),
    )
    last_login = Field(
        dtype=datetime,
        input_processor=DATE_PROCESSOR,
        serializer=serialize_date,
        parser=parse_date,
    )

    country = Field(dtype=str)
    region = Field(dtype=str)
    city = Field(dtype=str)

    external_link = Field(
        dtype=list,
        input_processor=URL_PROCESSOR,
        output_processor=_clear_list,
        serializer=JSON_SERIALIZER,
        parser=parse_json,
    )
    image_url = Field(
        dtype=list,
        input_processor=URL_PROCESSOR,
        output_processor=_clear_list,
        serializer=JSON_SERIALIZER,
        parser=parse_json,
    )
    image_file = Field(serializer=JSON_SERIALIZER, parser=parse_json)

    published_at = Field(
        dtype=datetime,
        input_processor=DATE_PROCESSOR,
        serializer=serialize_date,
        parser=parse_date,
    )
    updated_at = Field(
        dtype=datetime,
        input_processor=DATE_PROCESSOR,
        serializer=serialize_date,
        parser=parse_date,
    )
    scraped_at = Field(
        dtype=datetime,
        required=True,
        default=now,
        input_processor=DATE_PROCESSOR,
        serializer=serialize_date,
        parser=parse_date,
    )


class RatingItem(TypedItem):
    """ item representing a rating """

    JSON_OUTPUT = SETTINGS.get("FEED_FORMAT") in ("jl", "json", "jsonl", "jsonlines")
    BOOL_SERIALIZER = identity if JSON_OUTPUT else _serialize_bool

    item_id = Field(required=True, input_processor=IDENTITY)

    bgg_id = Field(dtype=int, dtype_convert=parse_int)
    bgg_user_name = Field(
        dtype=str, input_processor=MapCompose(identity, str, str.lower)
    )

    bgg_user_rating = Field(
        dtype=float,
        dtype_convert=parse_float,
        default=None,
        input_processor=POS_FLOAT_PROCESSOR,
    )
    bgg_user_owned = Field(
        dtype=bool,
        dtype_convert=parse_bool,
        default=None,
        input_processor=IDENTITY,
        serializer=BOOL_SERIALIZER,
        parser=parse_bool,
    )
    bgg_user_prev_owned = Field(
        dtype=bool,
        dtype_convert=parse_bool,
        default=None,
        input_processor=IDENTITY,
        serializer=BOOL_SERIALIZER,
        parser=parse_bool,
    )
    bgg_user_for_trade = Field(
        dtype=bool,
        dtype_convert=parse_bool,
        default=None,
        input_processor=IDENTITY,
        serializer=BOOL_SERIALIZER,
        parser=parse_bool,
    )
    bgg_user_want_in_trade = Field(
        dtype=bool,
        dtype_convert=parse_bool,
        default=None,
        input_processor=IDENTITY,
        serializer=BOOL_SERIALIZER,
        parser=parse_bool,
    )
    bgg_user_want_to_play = Field(
        dtype=bool,
        dtype_convert=parse_bool,
        default=None,
        input_processor=IDENTITY,
        serializer=BOOL_SERIALIZER,
        parser=parse_bool,
    )
    bgg_user_want_to_buy = Field(
        dtype=bool,
        dtype_convert=parse_bool,
        default=None,
        input_processor=IDENTITY,
        serializer=BOOL_SERIALIZER,
        parser=parse_bool,
    )
    bgg_user_preordered = Field(
        dtype=bool,
        dtype_convert=parse_bool,
        default=None,
        input_processor=IDENTITY,
        serializer=BOOL_SERIALIZER,
        parser=parse_bool,
    )
    bgg_user_wishlist = Field(
        dtype=int,
        dtype_convert=parse_int,
        input_processor=POS_INT_PROCESSOR,
        default=None,
    )
    bgg_user_play_count = Field(
        dtype=int, dtype_convert=parse_int, input_processor=NN_INT_PROCESSOR, default=0
    )

    bga_id = Field(dtype=str)
    bga_user_id = Field(dtype=str)
    bga_user_name = Field(dtype=str, input_processor=MapCompose(identity, str))
    bga_user_rating = Field(
        dtype=float,
        dtype_convert=parse_float,
        default=None,
        input_processor=POS_FLOAT_PROCESSOR,
    )

    comment = Field(
        dtype=str,
        input_processor=MapCompose(
            identity,
            str,
            remove_tags,
            replace_all_entities,
            partial(normalize_space, preserve_newline=True),
        ),
    )

    published_at = Field(
        dtype=datetime,
        input_processor=DATE_PROCESSOR,
        serializer=serialize_date,
        parser=parse_date,
    )
    updated_at = Field(
        dtype=datetime,
        input_processor=DATE_PROCESSOR,
        serializer=serialize_date,
        parser=parse_date,
    )
    scraped_at = Field(
        dtype=datetime,
        required=True,
        default=now,
        input_processor=DATE_PROCESSOR,
        serializer=serialize_date,
        parser=parse_date,
    )



# -*- coding: utf-8 -*-

"""Pull logs from Google Cloud PubSub queue."""

import argparse
import csv
import logging
import os
import sys

from itertools import count
from pathlib import Path
from time import sleep

from pytility import normalize_space

try:
    # pylint: disable=redefined-builtin
    from smart_open import open
except ImportError:
    pass

from .utils import now, pubsub_client

LOGGER = logging.getLogger(__name__)
BASE_DIR = Path(__file__).resolve().parent.parent


def _process_messages(messages, output, header=False, encoding="utf-8"):
    writer = csv.writer(output)

    if header:
        writer.writerow(("date", "user"))

    for message in messages:
        try:
            date = message.message.publish_time.replace(nanosecond=0).isoformat()
            user = normalize_space(message.message.data.decode(encoding)).lower()
            if date and user:
                writer.writerow((date, user))
                yield message.ack_id
            else:
                LOGGER.error("there was a problem processing message %r", message)

        except Exception:
            LOGGER.exception("unable to process message %r", message)


def _parse_args():
    parser = argparse.ArgumentParser(
        description="Pull logs from Google Cloud PubSub queue."
    )
    parser.add_argument(
        "--project",
        "-p",
        default=os.getenv("PULL_QUEUE_PROJECT"),
        help="Google Cloud project",
    )
    parser.add_argument(
        "--subscription",
        "-s",
        default=os.getenv("PULL_QUEUE_SUBSCRIPTION_LOGS"),
        help="Google Cloud PubSub subscription",
    )
    parser.add_argument(
        "--out-path",
        "-o",
        help="Output location",
    )
    parser.add_argument(
        "--header",
        "-H",
        action="store_true",
        help="include CSV header",
    )
    parser.add_argument(
        "--batch-size",
        "-b",
        type=int,
        default=1000,
        help="maximum batch size (max messages per pull and file)",
    )
    parser.add_argument(
        "--timeout",
        "-t",
        type=float,
        default=60,
        help="timeout for a pull in seconds",
    )
    parser.add_argument(
        "--sleep",
        "-S",
        type=float,
        help="sleep for that many seconds before start pulling",
    )
    parser.add_argument(
        "--no-ack",
        "-n",
        action="store_true",
        help="do not acknowledge messages to subscription",
    )
    parser.add_argument(
        "--verbose",
        "-v",
        action="count",
        default=0,
        help="log level (repeat for more verbosity)",
    )

    return parser.parse_args()


def main():
    """Pull logs from Google Cloud PubSub queue."""

    args = _parse_args()

    logging.basicConfig(
        stream=sys.stderr,
        level=logging.DEBUG if args.verbose > 0 else logging.INFO,
        format="%(asctime)s %(levelname)-8.8s [%(name)s:%(lineno)s] %(message)s",
    )

    LOGGER.info(args)

    if not args.project or not args.subscription:
        LOGGER.error("Google Cloud PubSub project and subscription are required")
        sys.exit(1)

    if args.sleep:
        LOGGER.info("going to sleep for %.1f seconds", args.sleep)
        sleep(args.sleep)

    client = pubsub_client()
    # pylint: disable=no-member
    subscription_path = client.subscription_path(args.project, args.subscription)

    for i in count():
        LOGGER.info("#%d: pulling subscription <%s>", i, subscription_path)

        try:
            response = client.pull(
                subscription=subscription_path,
                max_messages=args.batch_size,
                return_immediately=False,
                timeout=args.timeout,
            )
        except Exception:
            LOGGER.info("subscription <%s> timed out", subscription_path)
            break

        if not response or not response.received_messages:
            LOGGER.info(
                "nothing to be pulled from subscription <%s>", subscription_path
            )
            break

        if not args.out_path or args.out_path == "-":
            ack_ids = tuple(
                _process_messages(
                    messages=response.received_messages,
                    output=sys.stdout,
                    header=args.header and (i == 0),
                )
            )
        else:
            out_path = args.out_path.format(
                number=i, time=now().strftime("%Y-%m-%dT%H-%M-%S")
            )
            LOGGER.info("writing results to <%s>", out_path)
            with open(out_path, "w", newline="") as out_file:
                ack_ids = tuple(
                    _process_messages(
                        messages=response.received_messages,
                        output=out_file,
                        header=args.header,
                    )
                )

        LOGGER.info("%d message(s) succesfully processed", len(ack_ids))

        if ack_ids and not args.no_ack:
            LOGGER.info(
                "acknowledge %d message(s) in subscription <%s>",
                len(ack_ids),
                subscription_path,
            )
            client.acknowledge(subscription=subscription_path, ack_ids=ack_ids)

    LOGGER.info("Done.")


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""Command line entry point."""

import argparse
import logging
import os

from datetime import timezone
from pathlib import Path
from time import sleep
from shutil import rmtree

from pytility import normalize_space, parse_date
from scrapy.cmdline import execute
from scrapy.utils.job import job_dir as job_dir_from_settings
from scrapy.utils.log import configure_logging
from scrapy.utils.misc import arg_to_iter
from scrapy.utils.project import get_project_settings
from scrapy.utils.python import garbage_collect

from .utils import date_from_file, now

LOGGER = logging.getLogger(__name__)
DATE_FORMAT = "%Y-%m-%dT%H-%M-%S"
RESUMABLE_STATES = frozenset(("shutdown", "closespider_timeout"))


def _find_states(
    path_dir,
    state_file=".state",
    delete="finished",
    delete_non_state=False,
):
    path_dir = Path(path_dir).resolve()
    delete = frozenset(arg_to_iter(delete))
    result = {}

    if not path_dir.is_dir():
        LOGGER.warning("<%s> is not an existing dir", path_dir)
        return result

    LOGGER.info("Finding jobs and their states in <%s>", path_dir)

    for sub_dir in path_dir.iterdir():
        state_path = sub_dir / state_file

        if not sub_dir.is_dir() or not state_path.is_file():
            continue

        try:
            with state_path.open() as file_obj:
                state = normalize_space(next(file_obj, None))
        except Exception:
            LOGGER.exeception("Unable to read a state from <%s>", state_path)
            state = None

        if not state:
            LOGGER.warning("No valid state file in <%s>", sub_dir)

        if state in delete or (delete_non_state and not state):
            LOGGER.info("Deleting <%s> with state <%s>", sub_dir, state)
            rmtree(sub_dir, ignore_errors=True)
        elif state:
            result[sub_dir.name] = state

    return result


def _parse_args():
    parser = argparse.ArgumentParser(description="TODO")
    parser.add_argument("spider", help="TODO")
    parser.add_argument("--job-dir", "-j", help="TODO")
    parser.add_argument("--feeds-dir", "-f", help="TODO")
    parser.add_argument("--feeds-subdir", "-F", help="TODO")
    parser.add_argument(
        "--file-tag", "-t", default=os.getenv("SCRAPER_FILE_TAG"), help="TODO"
    )
    parser.add_argument("--dont-run-before", "-d", help="TODO")
    parser.add_argument(
        "--verbose",
        "-v",
        action="count",
        default=0,
        help="log level (repeat for more verbosity)",
    )

    return parser.parse_args()


def main():
    """Command line entry point."""

    settings = get_project_settings()
    configure_logging(settings)

    args = _parse_args()
    LOGGER.info(args)

    base_dir = Path(settings["BASE_DIR"]).resolve()
    cache_dir = base_dir / ".scrapy" / "httpcache"
    feeds_dir = Path(args.feeds_dir) if args.feeds_dir else base_dir / "feeds"
    feeds_dir = feeds_dir.resolve()
    feeds_dir_scraper = (
        feeds_dir / args.feeds_subdir if args.feeds_subdir else feeds_dir / args.spider
    )
    file_tag = normalize_space(args.file_tag)
    out_file = feeds_dir_scraper / "%(class)s" / f"%(time)s{file_tag}.jl"

    LOGGER.info("Output file will be <%s>", out_file)

    from_settings = job_dir_from_settings(settings)
    job_dir = (
        Path(args.job_dir)
        if args.job_dir
        else Path(from_settings)
        if from_settings
        else base_dir / "jobs" / args.spider
    )
    job_dir = job_dir.resolve()

    cache_dir.mkdir(parents=True, exist_ok=True)
    feeds_dir_scraper.mkdir(parents=True, exist_ok=True)
    job_dir.mkdir(parents=True, exist_ok=True)

    dont_run_before_file = job_dir / ".dont_run_before"
    dont_run_before = parse_date(
        args.dont_run_before, tzinfo=timezone.utc
    ) or date_from_file(dont_run_before_file, tzinfo=timezone.utc)

    if dont_run_before:
        LOGGER.info("Don't run before %s", dont_run_before.isoformat())
        sleep_seconds = dont_run_before.timestamp() - now().timestamp()
        if sleep_seconds > 0:
            LOGGER.info("Going to sleep for %.1f seconds", sleep_seconds)
            sleep(sleep_seconds)

    states = _find_states(
        job_dir, state_file=settings.get("STATE_TAG_FILE") or ".state"
    )

    running = sorted(sub_dir for sub_dir, state in states.items() if state == "running")

    if len(running) > 1:
        LOGGER.warning(
            "Found %d running jobs %s, please check and fix!", len(running), running
        )
        return

    if running:
        LOGGER.info("Found a running job <%s>, skipping...", running[0])
        return

    resumable = sorted(
        sub_dir for sub_dir, state in states.items() if state in RESUMABLE_STATES
    )

    if len(resumable) > 1:
        LOGGER.warning(
            "Found %d resumable jobs %s, please check and fix!",
            len(resumable),
            resumable,
        )
        return

    if resumable:
        LOGGER.info("Resuming previous job <%s>", resumable[0])

    job_tag = resumable[0] if resumable else now().strftime(DATE_FORMAT)
    curr_job = job_dir / job_tag

    command = [
        "scrapy",
        "crawl",
        args.spider,
        "--output",
        str(out_file),
        "--set",
        f"JOBDIR={curr_job}",
        "--set",
        f"DONT_RUN_BEFORE_FILE={dont_run_before_file}",
    ]

    try:
        execute(argv=command)
    finally:
        garbage_collect()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""BoardGameGeek GeekList spider."""

import re

from datetime import timezone

from pytility import parse_date, parse_int
from scrapy import Spider

from ..items import GameItem
from ..loaders import GameLoader
from ..utils import extract_bgg_id, now

TITLE_REGEX = re.compile(
    r"^\s*bgg\s*top.*from\s*(\d+\s*[a-z]+\s*\d+)\s*to\s*(\d+\s*[a-z]+\s*\d+).*$",
    re.IGNORECASE,
)


class BggGeekListSpider(Spider):
    """BoardGameGeek GeekList spider."""

    name = "bgg_geeklist"
    allowed_domains = ("boardgamegeek.com",)
    start_urls = (
        "https://www.boardgamegeek.com/geeklist/30543/bgg-top-50-statistics-meta-list",
    )
    item_classes = (GameItem,)

    # exclude Hall of Fame (197551) and The Thing from the Future (167330)
    # as they are not part of the rankings
    exclude_bgg_ids = frozenset((197551, 167330))

    custom_settings = {
        "DOWNLOAD_DELAY": 0.5,
        "CONCURRENT_REQUESTS_PER_DOMAIN": 8,
        "AUTOTHROTTLE_TARGET_CONCURRENCY": 4,
        "DELAYED_RETRY_ENABLED": True,
        "DELAYED_RETRY_HTTP_CODES": (202,),
        "DELAYED_RETRY_DELAY": 5.0,
        "AUTOTHROTTLE_HTTP_CODES": (429, 503, 504),
    }

    def parse_game(self, item, response, **kwargs):
        """Parse game."""

        title = item.css(".geeklist_item_title")

        for url in title.xpath(".//a/@href").extract():
            bgg_id = extract_bgg_id(response.urljoin(url))
            if bgg_id:
                break
        else:
            return None

        if bgg_id in self.exclude_bgg_ids:
            return None

        assert bgg_id

        ldr = GameLoader(
            item=GameItem(bgg_id=bgg_id, **kwargs),
            selector=item,
            response=response,
        )

        ldr.add_value("name", title.xpath(".//a[2]/text()").extract_first())
        ldr.add_xpath("image_url", ".//a/img[starts-with(@alt, 'Board Game:')]/@src")

        return ldr.load_item()

    def parse_geeklist(self, item, response, **kwargs):
        """Parse geeklist."""

        url = item.css(".geeklist_item_title").xpath(".//a[2]/@href").extract_first()

        if url and "/geeklist/" in url:
            return response.follow(url=url, callback=self.parse)

        return None

    def parse_item(self, item, response, **kwargs):
        """Decide on the item type and call corresponding method."""

        rank_text = (
            item.css(".geeklist_item_title").xpath(".//a[1]/text()").extract_first()
        )
        rank = parse_int(rank_text[:-1]) if rank_text else None

        return (
            self.parse_game(
                item=item,
                response=response,
                rank=rank,
                **kwargs,
            )
            or self.parse_geeklist(item=item, response=response)
        )

    def parse(self, response):
        """
        @url https://www.boardgamegeek.com/geeklist/30543/bgg-top-50-statistics-meta-list
        @returns items 0 0
        @returns requests 26
        """

        for next_page in response.xpath(
            "//a[contains(@title, 'page')]/@href"
        ).extract():
            yield response.follow(
                url=next_page,
                callback=self.parse,
            )

        scraped_at = now()

        for title in (
            response.xpath("/html/head/title/text()").extract()
            + response.css("div.geeklist_title::text").extract()
        ):
            match = TITLE_REGEX.match(title)
            published_at = (
                parse_date(match.group(2), tzinfo=timezone.utc) if match else None
            )
            if published_at:
                break
        else:
            published_at = None

        for item in response.xpath("//*[@data-objecttype = 'listitem']"):
            result = self.parse_item(
                item=item,
                response=response,
                published_at=published_at,
                scraped_at=scraped_at,
            )
            if result:
                yield result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Setup."""

# Template from https://github.com/navdeep-G/setup.py

import io
import os
import sys
from shutil import rmtree

from setuptools import find_packages, setup, Command

# Package meta-data.
NAME = "board-game-scraper"
DESCRIPTION = "Board games data scraping and processing from BoardGameGeek and more!"
KEYWORDS = (
    "board games",
    "tabletop games",
    "data",
    "datasets",
    "scraper",
    "scrapy",
    "spider",
    "boardgamegeek",
    "bgg",
    "ludoj",
    "ludoj-scraper",
)
URL_HOMEPAGE = "https://recommend.games/"
URL_DOCUMENTATION = (
    "https://gitlab.com/recommend.games/board-game-scraper/blob/master/README.md"
)
URL_FUNDING = "https://paypal.me/mschepke"
URL_THANKS = "https://saythanks.io/to/mk.schepke%40gmail.com"
URL_SOURCE = "https://gitlab.com/recommend.games/board-game-scraper"
URL_TRACKER = "https://gitlab.com/recommend.games/board-game-scraper/issues"
URL_EXAMPLES = None
URL_CHANGELOG = None
URL_MAILING = None
URL_TWITTER = "https://twitter.com/recommend_games"
EMAIL = "markus@recommend.games"
AUTHOR = "Markus Shepherd"
REQUIRES_PYTHON = ">=3.6.0"
VERSION = None  # will be read from __version__.py

# What packages are required for this module to be executed?
REQUIRED = (
    "awscli",
    "boto",
    "dedupe>=2.0.0",
    "google-cloud-pubsub",
    "jmespath",
    "pillow",
    "pympler",
    "pyspark",
    "python-dotenv",
    "pytility[dates]",
    "pytrie",
    "pyyaml",
    "requests",
    "scrapy<2.1.0",
    "scrapy-extensions",
    "twisted",
    "w3lib",
)

# What packages are optional?
EXTRAS = {
    "cloud": ("smart-open>=1.8.1",),
}

# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

HERE = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(HERE, "README.md"), encoding="utf-8") as f:
        LONG_DESCRIPTION = "\n" + f.read()
except FileNotFoundError:
    LONG_DESCRIPTION = DESCRIPTION

# Load the package's __version__.py module as a dictionary.
ABOUT = {}
if not VERSION:
    PROJECT_SLUG = NAME.lower().replace("-", "_").replace(" ", "_")
    with open(os.path.join(HERE, PROJECT_SLUG, "__version__.py")) as f:
        # pylint: disable=exec-used
        exec(f.read(), ABOUT)
else:
    ABOUT["__version__"] = VERSION


class UploadCommand(Command):
    """Support setup.py upload."""

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(string):
        """Prints things in bold."""
        print("\033[1m{0}\033[0m".format(string))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds…")
            rmtree(os.path.join(HERE, "dist"))
        except OSError:
            pass

        self.status("Building Source and Wheel (universal) distribution…")
        os.system("{0} setup.py sdist bdist_wheel --universal".format(sys.executable))

        self.status("Uploading the package to PyPI via Twine…")
        os.system("twine upload dist/*")

        self.status("Pushing git tags…")
        os.system("git tag v{0}".format(ABOUT["__version__"]))
        os.system("git push --tags")

        sys.exit()


# Where the magic happens:
setup(
    name=NAME,
    version=ABOUT["__version__"],
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    keywords=KEYWORDS,
    author=AUTHOR,
    author_email=EMAIL,
    url=URL_HOMEPAGE,
    project_urls={
        "Documentation": URL_DOCUMENTATION,
        "Funding": URL_FUNDING,
        "Say Thanks!": URL_THANKS,
        "Source": URL_SOURCE,
        "Tracker": URL_TRACKER,
        # "Examples": URL_EXAMPLES,
        # "Changelog": URL_CHANGELOG,
        # "Mailing List": URL_MAILING,
        "Twitter": URL_TWITTER,
    },
    python_requires=REQUIRES_PYTHON,
    packages=find_packages(exclude=("tests", "*.tests", "*.tests.*", "tests.*")),
    # If your package is a single module, use this instead of 'packages':
    # py_modules=(),
    entry_points={
        "console_scripts": (
            "bg-scraper=board_game_scraper.__main__:main",
            "bg-merge=board_game_scraper.merge:main",
            "bg-full-merge=board_game_scraper.full_merge:main",
            "bg-news=board_game_scraper.news:main",
            "bg-pull=board_game_scraper.pubsub_pull:main",
        ),
    },
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    include_package_data=True,
    license="MIT",
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Framework :: Scrapy",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Games/Entertainment :: Board Games",
    ],
    # $ setup.py publish support.
    cmdclass={"upload": UploadCommand},
)



# -*- coding: utf-8 -*-

""" Scrapy extensions """

import logging
import os

from datetime import timedelta, timezone
from pathlib import Path

from pytility import parse_date, parse_float
from scrapy import signals
from scrapy.exceptions import NotConfigured
from scrapy.utils.job import job_dir
from scrapy_extensions import LoopingExtension

from .utils import now, pubsub_client

LOGGER = logging.getLogger(__name__)


class PullQueueExtension(LoopingExtension):
    """ periodically pull from a queue """

    @classmethod
    def from_crawler(cls, crawler):
        """ init from crawler """

        if not crawler.settings.getbool("PULL_QUEUE_ENABLED"):
            raise NotConfigured

        project = crawler.settings.get("PULL_QUEUE_PROJECT")
        subscription = crawler.settings.get("PULL_QUEUE_SUBSCRIPTION")

        if not project or not subscription:
            raise NotConfigured

        interval = crawler.settings.getfloat("PULL_QUEUE_INTERVAL", 60 * 60)
        max_messages = crawler.settings.getint("PULL_QUEUE_MAX_MESSAGES", 100)
        prevent_rescrape_for = (
            crawler.settings.getfloat("PULL_QUEUE_PREVENT_RESCRAPE_FOR") or None
        )
        pull_timeout = crawler.settings.getfloat("PULL_QUEUE_PULL_TIMEOUT", 10)

        return cls(
            crawler=crawler,
            interval=interval,
            project=project,
            subscription=subscription,
            max_messages=max_messages,
            prevent_rescrape_for=prevent_rescrape_for,
            pull_timeout=pull_timeout,
        )

    def __init__(
        self,
        crawler,
        interval,
        project,
        subscription,
        max_messages=100,
        prevent_rescrape_for=None,
        pull_timeout=10,
    ):
        self.client = pubsub_client()

        if not self.client:
            LOGGER.error("Google Cloud Pub/Sub Client could not be initialised")
            raise NotConfigured

        # pylint: disable=no-member
        self.subscription_path = self.client.subscription_path(project, subscription)
        self.max_messages = max_messages

        prevent_rescrape_for = (
            prevent_rescrape_for
            if isinstance(prevent_rescrape_for, timedelta)
            else parse_float(prevent_rescrape_for)
        )
        self.prevent_rescrape_for = (
            timedelta(seconds=prevent_rescrape_for)
            if isinstance(prevent_rescrape_for, float)
            else prevent_rescrape_for
        )
        self.pull_timeout = parse_float(pull_timeout) or 10
        self.last_scraped = {}

        self.setup_looping_task(self._pull_queue, crawler, interval)

    def _pull_queue(self, spider):
        LOGGER.info("pulling subscription <%s>", self.subscription_path)

        try:
            # pylint: disable=no-member
            response = self.client.pull(
                subscription=self.subscription_path,
                max_messages=self.max_messages,
                return_immediately=False,
                timeout=self.pull_timeout,
            )
        except Exception:
            LOGGER.info("subscription <%s> timed out", self.subscription_path)
            return

        if not response or not response.received_messages:
            LOGGER.info(
                "nothing to be pulled from subscription <%s>", self.subscription_path
            )
            return

        ack_ids = [
            message.ack_id
            for message in response.received_messages
            if self.process_message(message.message, spider)
        ]

        if ack_ids:
            # pylint: disable=no-member
            self.client.acknowledge(
                subscription=self.subscription_path, ack_ids=ack_ids
            )

    # pylint: disable=no-self-use
    def process_message(self, message, spider, encoding="utf-8"):
        """ schedule collection request for user name """

        LOGGER.debug("processing message <%s>", message)

        user_name = message.data.decode(encoding)
        if not user_name or not hasattr(spider, "collection_request"):
            return True

        user_name = user_name.lower()

        if self.prevent_rescrape_for:
            last_scraped = self.last_scraped.get(user_name)
            curr_time = now()

            if last_scraped and last_scraped + self.prevent_rescrape_for > curr_time:
                LOGGER.info("dropped <%s>: last scraped %s", user_name, last_scraped)
                return True

            self.last_scraped[user_name] = curr_time

        LOGGER.info("scheduling collection request for <%s>", user_name)
        request = spider.collection_request(
            user_name=user_name, priority=1, dont_filter=True
        )
        spider.crawler.engine.crawl(request, spider)

        return True


class StateTag:
    """ writes a tag into JOBDIR with the state of the spider """

    @classmethod
    def from_crawler(cls, crawler):
        """ init from crawler """

        jobdir = job_dir(crawler.settings)

        if not jobdir:
            raise NotConfigured

        state_file = crawler.settings.get("STATE_TAG_FILE") or ".state"
        pid_file = crawler.settings.get("PID_TAG_FILE") or ".pid"

        obj = cls(jobdir, state_file, pid_file)

        crawler.signals.connect(obj._spider_opened, signals.spider_opened)
        crawler.signals.connect(obj._spider_closed, signals.spider_closed)

        return obj

    def __init__(self, jobdir, state_file, pid_file=None):
        os.makedirs(jobdir, exist_ok=True)
        self.state_path = os.path.join(jobdir, state_file)
        self.pid_file = os.path.join(jobdir, pid_file) if pid_file else None

    def _write(self, target, content):
        path = self.pid_file if target == "pid" else self.state_path

        if not path:
            return 0

        with open(path, "w") as out_file:
            return out_file.write(content)

    def _delete(self, target):
        path = self.pid_file if target == "pid" else self.state_path

        if not path:
            return False

        try:
            os.remove(path)
            return True

        except Exception as exc:
            LOGGER.exception(exc)

        return False

    def _spider_opened(self):
        self._write("state", "running")
        self._write("pid", str(os.getpid()))

    # pylint: disable=unused-argument
    def _spider_closed(self, spider, reason):
        self._write("state", reason)
        self._delete("pid")


class DontRunBeforeTag:
    """Writes a tag when to start the next run."""

    @classmethod
    def from_crawler(cls, crawler):
        """ init from crawler """

        tag_file = crawler.settings.get("DONT_RUN_BEFORE_FILE")
        date = parse_date(
            crawler.settings.get("DONT_RUN_BEFORE_DATE"), tzinfo=timezone.utc
        )
        seconds = crawler.settings.getfloat("DONT_RUN_BEFORE_SEC")

        if not tag_file or not (seconds or date):
            raise NotConfigured

        obj = cls(tag_file, date, seconds)

        crawler.signals.connect(obj._spider_opened, signals.spider_opened)

        return obj

    def __init__(self, tag_file, date=None, seconds=None):
        date = parse_date(date, tzinfo=timezone.utc)
        seconds = parse_float(seconds)

        if not date and not seconds:
            raise NotConfigured

        self.tag_file = Path(tag_file).resolve()
        self.tag_file.parent.mkdir(parents=True, exist_ok=True)
        self.date = date
        self.seconds = seconds

    def _spider_opened(self):
        date = self.date or now() + timedelta(seconds=self.seconds)

        LOGGER.info("Writing don't run before <%s> tag to <%s>", date, self.tag_file)

        with self.tag_file.open("w") as file_obj:
            file_obj.write(date.isoformat())
